#!/software/bin/perl
# ----------------------------------------------------------------------
#
# Author:        ts6
# Maintainer:    $Author: ajb $
# Created:       2008-02-05
# Last Modified: $Date: 2011-07-04 12:30:01 +0100 (Mon, 04 Jul 2011) $
# Id:            $Id: qseq2fastq.pl 13527 2011-07-04 11:30:01Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-28.1/bin/qseq2fastq.pl $
#
# ----------------------------------------------------------------------
#
# Convert a qseq file to a fastq file. Assumes scores in qseq are
# phred+64. Optionally output only PF or non-PF reads.
#
# ----------------------------------------------------------------------

use strict;
use warnings;
use English qw(-no_match_vars);
use Carp;
use Getopt::Long;
use FileHandle;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13527 $ =~ /(\d+)/mxs; $r; };

## no critic

sub process;
sub usage;
sub initialise;
sub timetag;

my $opts = initialise;

process;

exit;

# ----------------------------------------------------------------------
sub usage {

  ## no critic

  my ($short_name) = $PROGRAM_NAME =~ /([^\/]+)$/xm;

  print STDERR "\n";
  print STDERR "$short_name version $VERSION\n";
  print STDERR "\n";
  print STDERR "    options:\n";
  print STDERR "\n";
  print STDERR "    --filter               filter out non-PF reads\n";
  print STDERR "    --no-filter            output only non-PF reads\n";
  print STDERR "    --random <percentage>  randomly select specified percentage of reads\n";
  print STDERR "    --seed <seed>          random seed when randomly selecting\n";
  print STDERR "    --run <run id>         force run id (string) in read name (def: value from qseq file)\n";
  print STDERR "    --read <read id>       force read id (string) in read name (def: value from qseq file)\n";
  print STDERR "    --index                include index tag in read name\n";
  print STDERR "    --offset               add offset after x:y in read name\n";
  print STDERR "    --unique               check read names in each qseq file are unique\n";
  print STDERR "    --md5 <file>           run md5sum on the output, put result in this file\n";
  print STDERR "    --fastqcheck <file>    run fastqcheck on output, put result in this file\n";
  print STDERR "    --verbose              include step-by-step stats\n";
  print STDERR "    --help                 print this message and quit\n";
  print STDERR "\n";
  print STDERR "    Output to stdout, default is to output all reads.\n";
  print STDERR "\n";

}

# ----------------------------------------------------------------------
sub process {

  my $count_in  = 0;
  my $count_out = 0;

  printf STDERR "%s%d qseq2fastq.pl starting...\n", timetag, $PID, $count_in, $count_out;

  my $do_filter = exists $opts->{'filter'};
  my $no_filter = exists $opts->{'no-filter'};

  my $md5_fh;
  if (exists $opts->{'md5'}) {
    $md5_fh = FileHandle->new();
    $md5_fh->open("| md5sum > $opts->{md5}") or die "Failed to open md5 file $opts->{md5}: $!\n";
  }

  my $fastqcheck_fh;
  if (exists $opts->{'fastqcheck'}) {
    $fastqcheck_fh = FileHandle->new();
    $fastqcheck_fh->open("| fastqcheck > $opts->{fastqcheck}") or die "Failed to open fastqcheck file $opts->{fastqcheck}: $!\n";
  }

  if (exists $opts->{'seed'}) {
    srand ($opts->{'seed'});
  }

  my $file = "";
  my $offset = 0;
  my %names = ();

  while (my $line = <>) {    # sets $line

    if( $ARGV ne $file ){
      $file = $ARGV;
      printf STDERR "Reading $file\n" if $opts->{'verbose'};
      $offset=0;
      %names = ();
    }else{
      $offset++;
    }

    chomp $line;
    ++$count_in;

    my ($instr, $run_id, $lane, $tile, $x, $y, $index, $read, $bases, $quals, $filter)
        = split /\t/, $line;

    if (exists $opts->{'run'}) {
      $run_id = $opts->{'run'};    # use command line parm if present
    }

    if (exists $opts->{'read'}) {
      $read = $opts->{'read'};    # use command line parm if present
    }

    my $name = "\@${instr}_$run_id:$lane:$tile:$x:$y";
    $name .= ":$offset" if exists $opts->{'offset'};
    $name .= "\#$index" if exists $opts->{'index'};
    $name .= "/$read";

    if (exists $opts->{'unique'}){
      my $key = "$x:$y";
      die "Duplicate read name $name\n" if exists $names{$key};
      $names{$key} = 1;
    }

    next if (exists $opts->{'random'} and rand (100) > $opts->{'random'});

    if ( ($filter and ! $no_filter) or ( ! $filter and ! $do_filter) ) {

      $bases =~ tr/./N/;               # turn dots into Ns

      ##### Commented out stuff below does logs-odds to phred conversion.
      ####$quals =~ tr/!-\175/!!!!!!!!!!!!!!!!!!!!!!""""""##$$%%&&-++,-\136/;
      $quals =~ tr/!-\175/!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!-\136/;

      my $output = "$name\n$bases\n+\n$quals\n";

      print $output;
      $md5_fh->print($output) if $opts->{'md5'};
      $fastqcheck_fh->print($output) if $opts->{'fastqcheck'};

      ++$count_out;

    }

  }

  $md5_fh->close() if $opts->{'md5'};
  $fastqcheck_fh->close() if $opts->{'fastqcheck'};

  printf STDERR "%s%d qseq2fastq.pl read %d reads, wrote %d\n", timetag, $PID, $count_in, $count_out;

}

# ----------------------------------------------------------------------
sub initialise {

  my %opts;
  my $rc = GetOptions(\%opts, 'help', 'verbose', 'filter', 'no-filter', 'random=i', 'seed=i', 'run=i', 'read=s', 'index', 'offset', 'unique', 'md5=s', 'fastqcheck=s');
  if ( ! $rc) {
    print {*STDERR} "\nerror in command line parameters\n" or croak 'print failed';
    usage;
    exit 1;
  }

  if (exists $opts{'help'}) {
    usage;
    exit;
  }

  if (exists $opts{'filter'} and exists $opts{'no-filter'}) {
    print {*STDERR} "\nERROR: --filter and --no-filter are mutually exclusive.\n" or croak 'print failed';
    usage;
    exit 1;
  }

  return \%opts;

}

# ----------------------------------------------------------------------
sub timetag {

  my ($s, $m, $h, $D, $M, $Y) = localtime (time());
  return (sprintf "%04d%02d%02d-%02d%02d%02d ", $Y+1900, $M+1, $D, $h, $m, $s);

}
