#############
# $Id: pluggable.pm 13527 2011-07-04 11:30:01Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-06-11
# Last Changed On: $Date: 2011-07-04 12:30:01 +0100 (Mon, 04 Jul 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-28.1/lib/npg_pipeline/pluggable.pm $

package npg_pipeline::pluggable;
use Moose;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13527 $ =~ /(\d+)/mxs; $r; };
use npg_common::mailer;

use npg::api::run;
use npg_pipeline::dispatch_tree;

extends q{npg_pipeline::base};

with qw{npg_common::roles::run::path_info};
with qw{npg_common::roles::run::long_info};
with qw{npg_pipeline::roles::generic::pipeline
        npg_pipeline::roles::generic::other_params
        npg_pipeline::roles::business::flag_options
  };

has q{analysis_type}      => (isa => q{Str}, is => q{ro});
has q{alert}              => (isa => q{Str}, is => q{ro}, writer => q{_set_alert});

##### privately stored information

sub fields {
  return qw{
    id_run lane script_name alert other_params
  };
}

sub main {
  my ($self) = @_;
  $self->log(q{You need a sub class of this in your schema module});
  return 1;
}

################
# mail people who are listed in alert
sub _mailer {
  my ($self, $arg_refs) = @_;
  $arg_refs ||= {};

  if (!$arg_refs->{to}) {
    my $alert = $self->alert();

    if ($alert !~ /@/xms) {
      $alert .= q{@}.q{sanger.ac.uk};
      $self->_set_alert($alert);
    }
    $arg_refs->{to} = $alert;
  }

  $arg_refs->{from}    ||= $self->mail_from();
  $arg_refs->{subject} ||= q{Functional script }.$self->script_name().q{ schema: }.__PACKAGE__;
  if (!$arg_refs->{body}) {
    $arg_refs->{body} = qq{Parameters:\n};
    foreach my $f ($self->fields()) {
      if ($self->$f()) {
        $arg_refs->{body} .= qq{$f: }.$self->$f().qq{\n};
      }
    }
    if (ref$self->all_job_ids() && scalar@{$self->all_job_ids()}) {
      $arg_refs->{body} .= qq{\nJobs submitted:\n};
      foreach my $jid (@{$self->all_job_ids()}) {
        $arg_refs->{body} .= qq{\t$jid\n};
      }
    }

    $arg_refs->{body} .= qq{\nCheers\n\nNew Pipeline Development\n\n}.$self->memory_log();
  }

  eval {
    npg_common::mailer->new($arg_refs)->mail();
  } or do {
    croak $EVAL_ERROR;
  };
  return 1;
}

no Moose;
__PACKAGE__->meta->make_immutable;

1;
__END__

=head1 NAME

  npg_pipeline::pluggable

=head1 VERSION

$LastChangedRevision: 13527 $

=head1 SYNOPSIS

  my $oDerived = npg_pipeline::pluggable::<derived_class>->new($arg_ref_hash);

=head1 DESCRIPTION

This is a superclass to a pluggable::schema module which should be created to run your schema. The objective
is that you will create a 'flag_waver' pluggable module which will control running functions in a determined
array order, but that those functions can (in theory) be placed in whichever order the user chooses.

  package npg_pipeline::pluggable::schema;
  use Moose;
  use Carp;
  use English qw{-no_match_vars};
  extends qw{npg_common::pluggable};

  sub main {
    my ($self, @args) = @_;
    # code to work through your array of functions
  }
  ### Your additional functions go here, or could be added to superclass if they are to be used in more than one schema
  no Moose;
  __package__->meta->make_immutable;
  1;

In the main method, you should define how to work through the given array (or you could define your array here and save passing it through).
You could also choose how to parallelise some functions should that be required (e.g. jobs being submitted to LSF that do not depend on each other)

e.g.

Straight Top to Bottom (loop through array of functions, passing the jobs deps from one function to the next):

  sub main {
    my ($self, @args) = @_;
    my @functions_to_run_in_order = ref$self->function_order() eq q{ARRAY}  ? @{$self->function_order}
                                  :                                           (qw{a predetermined function order with finish})
                                  ;

    my @job_ids; # an array to hold LSF job ids from the previous level to pass down
    my $all_job_ids = $self->all_job_ids(); # internal store of all job ids from jobs submitted to LSF

    foreach my $function (@functions_to_run_in_order) {
      $self->log(q{****** Processing }.$function.q{ ******}); # logging function inherited from pluggable.pm
      my $job_deps = $self->lsf_job_complete_requirements(@job_ids); # inherited from pluggable.pm, converts any array of job ids (including array of the direct returned strings from lsf) into a job requirement string for LSF submission
      $self->log(qq{$job_deps for $function});
      unshift @args, $job_reqs; preppends the job_reqs to the args array, so that functions cn obtain it as the first item of the array
      @job_ids = $self->$function(@args);
      push @{$all_job_ids}, @job_ids;
      $self->dispatch_tree->append_to_functions({  # dispatch tree inherited from pluggable.pm, and this logs how the tree is being formed, so that it could be retrieved later
        function => $function,
        job_ids_launched => \@job_ids,
        job_dependencies => $job_deps,
      });
    }
    return 1;
  }

With Parallelisable Functions (loop though the array of functions, but if some are able to be submitted at the same time, then parallelise them):

  sub main {
    my ($self, @args) = @_;
    my @functions_to_run_in_order = ref$self->function_order() eq q{ARRAY}  ? @{$self->function_order}
                                  :                                           (qw{a predetermined function order})
                                  ;

    my %parallelise_functions     = ref$self->parallelise() eq q{HASH}      ? %{$self->parallelise()}
                                  :                                           ( a => { funa => 1, funb => 1 }, b => { fund => 1, fune => 1 } )
                                  ;

    my @job_ids; # an array to hold LSF job ids from the previous level to pass down
    my $all_job_ids = $self->all_job_ids(); # internal store of all job ids from jobs submitted to LSF
    my @hold_parallel_ids; # temporary store of LSF job ids where scripts can be parallelised through LSF, before being made as the dependencies on the next level

    ##############
    # run through the selected functions in order selected

    my $add_to_hold;
    foreach my $function (@functions_to_run_in_order) {
      $self->log(q{****** Processing }.$function.q{ ******}); # logging function inherited from pluggable.pm
      my $job_reqs = $self->lsf_job_complete_requirements(@job_ids); # inherited from pluggable.pm, converts any array of job ids (including array of the direct returned strings from lsf) into a job requirement string for LSF submission
      $self->log(qq{$job_reqs for $function});
      unshift @args, $job_reqs; preppends the job_reqs to the args array, so that functions cn obtain it as the first item of the array

      ##############
      # if the previous function was parallelisable
      my $parallel_done;
      if ($add_to_hold) {

        my $parallelise_string = q{};
        foreach my $function (sort keys %{$parallelise_functions{$add_to_hold}}) {
          $parallelise_string .= qq{$function }
        }

        ###############
        # check if this function is also part of the same parallelisable group
        if($parallelise_functions{$add_to_hold}->{$function}) {
          ###############
          # if so, then run, and capture the job ids in temporary holding array
          $self->log(qq{Able to parallelise $parallelise_string});
          my $job_deps = $args[0];
          my @jids = $self->$function(@args);
          push @hold_parallel_ids, @jids;
          $self->dispatch_tree->append_to_functions({ function => $function,
              job_ids_launched => \@jids,
              job_dependencies => $job_deps,
            }); # dispatch tree inherited from pluggable.pm, and this logs how the tree is being formed, so that it could be retrieved later
          $parallel_done++;
        } else { # if not, then set up that we are moving to the next level off the job tree
          $self->log(qq{$function not parallelisable with $parallelise_string\n\tremoving parallelisable status and setting job id requirements for next level});
          $add_to_hold = q{};
          @job_ids = @hold_parallel_ids;
          push @{$all_job_ids}, @job_ids;
          $job_reqs = $self->lsf_job_complete_requirements(@job_ids);
          $self->log(qq{$job_reqs for $function});
          shift@args;
          unshift @args, $job_reqs;
          @hold_parallel_ids = ();
        }
      }
      if (!$parallel_done) {
        my $done;
        #############
        # see if this function is possibly parallelisable, and if so, dispatch as such
        foreach my $key (sort keys%parallelise_functions) {
          if($parallelise_functions{$key}->{$function}) {
            $add_to_hold = $key;
            my $parallelise_string = q{};
            foreach my $function (sort keys %{$parallelise_functions{$add_to_hold}}) {
              $parallelise_string .= qq{$function }
            }
            $self->log(qq{Able to parallelise $parallelise_string});
            my $job_deps = $args[0];
            my @jids = $self->$function(@args);
            push @hold_parallel_ids, @jids;
            $self->dispatch_tree->append_to_functions({ function => $function,
              job_ids_launched => \@jids,
              job_dependencies => $job_deps,
            });
            $done++;
            last;
          }
        }

        if (!$done) { # If through to here, then the jobs can just be dispatched as their own level
          my $job_deps = $args[0];
          @job_ids = $self->$function(@args);
          push @{$all_job_ids}, @job_ids;
          $self->dispatch_tree->append_to_functions({ function => $function,
            job_ids_launched => \@job_ids,
            job_dependencies => $job_deps,
          });
        }
      }
    }

    return 1;
  }

You can then define your functions/methods to run. These should be looking to call out to another piece of code (or object) which then
performs any tasks, returning only the dependencies anything further might need to run (LSF job ids, etc)

sub funa {
  my ($self, @args) = @_;
  my $job_dependencies = shift @args;
  require an::object;
  my $obj = an::object->new({
    run_folder => $self->run_folder(),
    id_run => $self->id_run(),
    job_dependencies => $job_dependencies,
  });
  my @job_ids = $obj->code_method_to_run();
  return @job_ids;
}

Anything that is required of the code to run should be dealt with in the object itself. This 'flag_waver' pluggable module should not be dealing
with extra code. It is up to code in the object to decide how, and on what it should run, and if there are dependencies, what those are, outside
of if submitting to LSF, what you have passed through.

You have the option of calling an internal _mailer method passing parameters to email when the script is running wiht a particular function/method

  $self->_mailer({
    to => $sEmail # will default to using the value of the alert parameter, but will croak if nothing available
    from => $sEmail # will default to $npg_common::pluggable::DEFAULT_MAIL_FROM
    subject => $sSubject # will default to a string with the script name and schema used
    body => $sBody # will default to cycling through most of the parameters
  });



=head1 SUBROUTINES/METHODS

=head2 new - instantiates a new object

  my $oDerivedClass = npg_pipeline::pluggable::<derived_class(schema_name)>->new({
    id_run => $iIDRun, # optional parameter offered by default
    lane   => $iLane, # optional parameter offered by default
    alert => $email, # optional, but needed if you want to use the _mailer method without hard_coding in the to parameter
    other_params => $sOtherParams, # stringify parameters you need, and then in your code, break it down
  });

=head2 BUILD - on creation, sets the time and timestamp attributes

=head2 fields - returns an array of the inbuilt accessors for parameters

=head2 main - method to call which should be subclassed within the schema module

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item strict

=item warnings

=item Carp

=item English -no_match_vars

=item Readonly

=item base

=item Class::Accessor

=item npg_common::mailer

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
