#############
# $Id: Spidering.pm 13527 2011-07-04 11:30:01Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-06-15
# Last Changed On: $Date: 2011-07-04 12:30:01 +0100 (Mon, 04 Jul 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-28.1/lib/npg_pipeline/roles/business/Spidering.pm $

package npg_pipeline::roles::business::Spidering;
use Moose::Role;
use Carp;
use English qw{-no_match_vars};
use List::MoreUtils qw(none);
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13527 $ =~ /(\d+)/mxs; $r; };
use npg_common::request;

with qw{npg_common::roles::test_type_of_value};

=head1 NAME

npg_pipeline::roles::business::Spidering

=head1 VERSION

$LastChangedRevision: 13527 $

=head1 SYNOPSIS

  package My:Spidering::Packaage;
  use Moose;
  with qw{npg_pipeline::roles::business::Spidering};

=head1 DESCRIPTION

Provides options and intelligence for spidering out of information.
The objective of which is to get the data discovered in a serialised
way and then presumably, if the option is there, get it cached, so
that when the pipeline wants it again, it is retrieved from any cache
files.

=head1 SUBROUTINES/METHODS

=head2 save2cache

boolean attribute which tells constructs an npg::api::util object for objects that tell it to save the data it retrieves to
a cache on the filesystem, which is dependent on $ENV{NPG_WEBSERVICE_CACHE_DIR} having an real directory path in it

=cut

has q{save2cache} => (
  isa => q{Bool},
  is => q{ro},
  predicate => q{has_save2cache},
);

=head2 retrieve_all_npg

Spiders out and retrieves all npg related information given a npg::api class name and it's primary key
It returns the initial top level object, or croaks at any given time.

=cut

sub retrieve_all_npg {
  my ( $self, $primary_key, $class ) = @_;

  eval { require $self->convert_package_name_to_filename( $class ); } or do { croak $EVAL_ERROR; };

  my $spidering_methods = $self->_grouped_spidering_methods( q{npg} );

  my $primary_key_name = $self->_primary_key_name( $class );

  my $attrs = {
    $primary_key_name => $primary_key,
  };
  # we may need to add a special useragent, in which case, this will do so
  $self->_add_repository_and_useragent( {
    href => $attrs,
    util => 1,
  } );

  my $object;

  eval {

    # npg run_status_dicts are almost always likely to be needed somewhere
    require npg::api::run_status_dict;
    my $rsd = npg::api::run_status_dict->new( $attrs );
    $self->_loop_methods( $rsd, $spidering_methods );

    $object = $class->new( $attrs );
    $self->_loop_methods( $object, $spidering_methods );
    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return $object;
}

=head2 retrieve_all_sequencescape

Spiders out and retrieves all st related information given a st::api class name and it's primary key
It returns the initial top level object, or croaks at any given time.

=cut

sub retrieve_all_sequencescape {
  my ( $self, $primary_key, $class ) = @_;

  eval { require $self->convert_package_name_to_filename( $class ); } or do { croak $EVAL_ERROR; };

  my $spidering_methods = $self->_grouped_spidering_methods( q{st} );

  my $primary_key_name = $self->_primary_key_name( $class );

  my $attrs = {
    $primary_key_name => $primary_key,
  };

  if ( $self->has_useragent() ) {
    $attrs->{useragent} = $self->useragent();
  }

  $self->_add_repository_and_useragent( {
    href => $attrs,
    util => 1,
  } );

  my $object;

  eval {
    $object = $class->new( $attrs );
    $self->_loop_methods( $object, $spidering_methods );
    1;
  } or do {
    croak $EVAL_ERROR;
  };
  return $object;
}

=head2 convert_package_name_to_filename

converts a package name (My::Package:Name) to the corresponding filename (My/Package/Name.pm)
for use in require

  my $filename = $self->convert_package_name_to_filename( $package_name );

=cut

sub convert_package_name_to_filename {
  my ( $self, $package ) = @_;

  $package =~ s{::}{/}gxms;
  if ( $package !~ /[.]pm\z/xms ) {
    $package .= q{.pm};
  }

  return $package;
}

=head2 repository

attribute which can set and override the base root for finding the repository where references and adapters are kept
don't set this unless you really mean it

=cut

has q{repository} => (
  isa => q{Str},
  is => q{ro},
  predicate => q{has_repository},
  documentation => q{Only set this if you want the reference repository root to be different from the standard},
);

=head2 useragent
=head2 npg_api_util

If you need to do some special things, then you can set pass in these objects. An npg_api_util object will be created if save2cache
has been set to true, or if you have passed in a useragent

Be warned, you would do well to leave these alone if you don't know why you should set them (and possibly, even if you do :) )

=cut

# useragent and util - should only be used if needed
has q{useragent} => (
  isa => q{Object},
  is  => q{rw},
  predicate => q{has_useragent},
);

has q{npg_api_util} => (
 isa => q{Object},
 is => q{rw},
 lazy_build => 1,
);

###########
# private methods

# convert a class name to a method key in order to locate the methods we
# should want to run on that class in order to locate the correct info
sub _convert_class_to_method_key {
  my ( $self, $class ) = @_;
  $class =~ s/::/_/gxms;
  return $class;
}

# generally, the primary key field is the first one of the fields list, but if there is no list,
# then create most likely name, which is the format 
sub _primary_key_name {
  my ( $self, $class ) = @_;

  my $primary_key_name;
  my @fields = $class->fields();
  if ( $fields[0] ) {
    $primary_key_name = shift @fields;
  } else {
    my @class_split = split /::/xms, $class;
    my $type = pop @class_split;
    $primary_key_name = $type . q{_id};
  }

  return $primary_key_name;
}

# we need to loop over objects to retrieve the correct information
# _loop_methods takes the object and the hashref of spidering methods
# so it can determine what is needed for that object.
# Also launches itself again if any results are further objects
sub _loop_methods {
  my ( $self, $object, $spidering_methods ) = @_;

  if ( ! $self->is_object( $object ) ) {
    return;
  }

  my $method_key = $self->_convert_class_to_method_key( ref $object );
  $self->log( q{Looping methods for } . ref $object );

  if ( ! $spidering_methods->{$method_key} ) {
    $self->log( q{No methods to loop here - will drop back} );
    return;
  }

  if ( ref $object eq q{npg::api::run_lane} ) {
    $self->_get_references( $object );
  }

  # multiplex lanes use a different route to obtain insert sizes, so work through the pool description to obtain them
  if ( ref $object eq q{st::api::lane} ) {
    my $pool_desc = $object->pool_description();
    if ( $pool_desc ) {
      foreach my $index ( keys %{ $pool_desc } ) {
        my $lib_tube_asset_id = $pool_desc->{$index}->{lib_tube_asset_id};
	if ( defined $lib_tube_asset_id ) {
	  my $lib_tube = st::api::asset->new({
	    util => $object->util(),
	    useragent => $object->useragent(),
	    id => $lib_tube_asset_id,
	  });
	  if ( ! $self->_already_seen( $lib_tube ) ) {
            $self->_loop_methods( $lib_tube, $spidering_methods );
          }
	}
      }
    }
  }

  foreach my $method ( keys %{ $spidering_methods->{$method_key} } ) {
    $self->log( qq{\tMethod: $method} );
    eval {
      my @return = $object->$method();
      if ( ref $return[0] eq q{ARRAY} ) {
        @return = @{ $return[0] };
      }

      foreach my $next_object ( @return ) {
        if ( $self->is_object( $next_object ) ) {
          if ( ! $self->_already_seen( $next_object ) ) {
            $self->_loop_methods( $next_object, $spidering_methods );
          }
        }
      }
      1;
    } or do {
      if ( $method ne q{run_pair} ) {
        croak $EVAL_ERROR;
      }
    };
  }

  return;

}

# retrieve the hashref from spidering_methods conf file, and filter it
# based on the filter provided, which should be npg or st (the first part
# of the class name)
sub _grouped_spidering_methods {
  my ( $self, $filter ) = @_;

  my %spidering_methods = %{ $self->spidering_methods_conf() };

  foreach my $method ( keys %spidering_methods ) {
    if ( $method !~ m/\A$filter/ixms ) {
      delete $spidering_methods{$method};
    }
  }
  return \%spidering_methods;
}

# some objects will call back to objects already retrieved once. We don't want to infinitely call the same stuff over and over
has q{_seen_that_already} => (
  isa => q{HashRef},
  is => q{rw},
  clearer => q{clear_seen_hash},
  lazy_build => 1,
);

sub _build__seen_that_already {
  my ( $self ) = @_;
  return {};
}

sub _already_seen {
  my ( $self, $object ) = @_;

  return 0 if ! $object;

  my $seen = $self->_seen_that_already();

  # for most objects with a fields array, the first item is a primary key accessor
  if ( $object->can( q{fields} ) ) {

    my @fields = $object->fields();
    my $primary_key = shift @fields;

    my $pk_value = $object->$primary_key();
    my $class = ref $object;
    if ( ! $pk_value ) {
      return 0;
    }
    if ( $seen->{$class}->{$pk_value} ) {
      return 1;
    }

    $seen->{$class}->{$pk_value}++;
  }

  return 0;
}

sub _build_npg_api_util {
  my ( $self ) = @_;

  my $attrs = {};

  if ( $self->has_useragent() ) {
    $attrs->{useragent} = $self->useragent();
  }

  if ( $self->has_save2cache() ) {
    $attrs->{save2cache} = $self->save2cache();
  }

  require npg::api::util;
  return npg::api::util->new( $attrs );
}


# ensure any calls made include looking for stuff to help obtain reference paths

sub _get_references {
  my ( $self, $run_lane ) = @_;

  require npg_common::sequence::reference;
  my $attrs = {
    id_run => $run_lane->id_run(),
    position => $run_lane->position(),
  };

  # this is probably likely to be used mainly for testing, although you could do a bit of overriding here
  $self->_add_repository_and_useragent( {
    href => $attrs,
    api_util => 1,
    repository => 1,
  } );

  my $ref;
  eval {
    $self->log( qq{\treference_finder for position:} . $run_lane->position() );
    $ref = npg_common::sequence::reference->new( $attrs );
    $ref->refs();
    foreach my $msg ( $ref->messages()->messages() ) {
      if ( $self->can( q{log} ) ) {
        $self->log( $msg );
      } else {
        carp $msg;
      }
    }
    1;
  } or do {
    if ( $self->can( q{log} ) ) {
      $self->log( $EVAL_ERROR );
    } else {
      carp $EVAL_ERROR;
    }
    croak 'Error during attempt to get references';
  }; # don't worry to much here at the moment just make sure that messages from reference finder come back to us

  return 1;
}

# if you want to add your own repository and/or useragent to a href
# this is quite likely if you have set save2cache, as it is probably the only way to get it through
sub _add_repository_and_useragent {
  my ( $self, $arg_refs ) = @_;

  if ( $arg_refs->{util} || $arg_refs->{api_util} ) {
    if ( $self->save2cache() || $self->has_useragent() || $ENV{npg_common::request::cache_dir_var_name()} ) {
      if ( $arg_refs->{api_util} ) {
        $arg_refs->{href}->{api_util} = $self->npg_api_util();
      }
      if ( $arg_refs->{util} ) {
        $arg_refs->{href}->{util} = $self->npg_api_util();
      }
    }
  }

  if ($arg_refs->{repository} ) {
    if ( $self->has_repository() ) {
      $arg_refs->{href}->{repository} = $self->repository();
    }
  }

  return 1;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=item List::MoreUtils

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
