#############
# $Id: pipeline.pm 13527 2011-07-04 11:30:01Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-11-02
# Last Changed On: $Date: 2011-07-04 12:30:01 +0100 (Mon, 04 Jul 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-28.1/lib/npg_pipeline/roles/generic/pipeline.pm $

package npg_pipeline::roles::generic::pipeline;
use Moose::Role;
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13527 $ =~ /(\d+)/mxs; $r; };
use npg_common::request;

requires qw{log};

=head1 NAME

npg_pipeline::roles::generic::pipeline

=head1 VERSION

$LastChangedRevision: 13527 $

=head1 SYNOPSIS

  package MyPackage;
  use Moose;
  ...
  with qw{npg_pipeline::roles::generic::pipeline};

=head1 DESCRIPTION

A role providing some generic pluggable pipeline methods. This role requires a log method to be provided for the finish method to be able to do some cleanup

=head1 SUBROUTINES/METHODS

=head2 function_order - unless you want to use a predetermined function order, provide an arrayref containing it

  my $aFunctionOrder = $class->function_order();

=cut

has q{function_order} => (isa => q{ArrayRef}, is => q{ro}, documentation => q{unless you want to use a predetermined function order, provide an arrayref containing it});

=head2 finish - last method to be called in your arrayref of functions. Handles writing some final information to the log, plus
  if you have an alert method and an mailer method, runs that
  if you have used the dispatch_tree object, stored in a method dispatch, will append the JSON version to the log

  $aFunctionOrder = [qw(
     some
     functions
     in
     the
     order
     wanted
     finish
  )];

=cut

sub finish {
  my ($self, @args) = @_;

  $self->log('finish');
  $self->log('I have finished creating the lsf jobs');
  $self->log(q{Total LSF jobs submitted: }.scalar@{$self->all_job_ids()});

  $self->log(qq{$args[0]});
  if ($self->can(q{dispatch_tree}) && $self->has_dispatch_tree()) {
    $self->log(qq{JSON Dispatch Tree:\n\n}.$self->dispatch_tree()->tree_as_json().qq{\n\n});
  }

  if ($self->can(q{alert}) && $self->alert() && $self->can(q{_mailer})) {
    $self->_mailer();
  }

  my $cache_dir_var_name = npg_common::request->cache_dir_var_name();
  $self->log( q[Unsetting $] .q[ENV{] . $cache_dir_var_name . q[}] );
  $ENV{$cache_dir_var_name} = q{}; ## no critic (Variables::RequireLocalizedPunctuationVars)

  return ();
}

=head2 lsf_job_complete_requirements - takes an array of job ids, and returns a -w'done(X) && done(Y) ....' string for use in lsf job submissions

  my $sLSFJobCompleteRequirements = $class->lsf_job_complete_requirements(@JobIds);

note, this can just take the string returned from a bsub job, as it parses for the number in the string

=cut

sub lsf_job_complete_requirements {
  my ($self, @job_ids) = @_;

  if (scalar@job_ids == 0) {
    return q{}; # jobs won't submit with -w'', so needs to be an empty string if there are no dependencies
  }

  foreach my $job_id (@job_ids) {
    # bsub normally returns a human readable submission, so just get the job id
    ($job_id) = $job_id =~ /(\d+)/xms;
    # convert to a done requirement
    $job_id = qq{done($job_id)};
  }

  # return the string which will be inserted as the requirement option
  return q{-w'}.(join q{ && }, @job_ids).q{'};
}

=head2 all_job_ids - returns an arrayref of all the job ids submitted, do not set on construction

  my $aAllJobIds = $class->all_job_ids();

=cut

has q{all_job_ids}  => (isa => q{ArrayRef}, is => q{ro}, lazy_build => 1,
                        documentation => q{Do not set on Construction of object},);

sub _build_all_job_ids {
  return [];
}

=head2 parallelise - store and return a hashref of functions which can be submitted to LSF and run at the same time

  my $hParallelise = {
    a => {func_a => 1, func_b => 1,},
    b => {func_g => 1, func_h => 1, func_i => 1,},
  };
  my $class_object = __PACKAGE__->new({
    parallelise => $hParallelise,
  });

This would mean that func_a and func_b would be submitted to LSF with no dependency on each other, and likewise func_g, func_h and func_i

  my $hParallelise = $class->parallelise(); 

If you provide a _build_parallelise method in your class, you can avoid needing to set this on construction, and code a parallelise hash within it
(eg, you know certain functions can always be submitted in parallel, so take it away from user decision)

=cut

has q{parallelise}  => (isa => q{HashRef}, is => q{ro}, lazy_build => 1);

1;
__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
